;;;; mother modular generic compiler project
; (c)2002/GPL Oskar Schirmer <oskar@scara.com>
; see file COPYING for GPL license details

; eval.scm: evaluator
; usage:
;   eval predicate-description-file <input-file >output-file
;
; the predicate-description-file contains a list of predicate
; definitions.
; it is a semantic extension to the extended-affixes-grammar.
;
; predicates: (rules)
; rule: (predicate-name ((formalparameters)) .
;        (list of elements, maybe empty))
; element: ((actualparameters) predicate-name)
; parameters: list of in and out: ((in-params) (out-params))
;
; parameter: (list of expressions)
; expression: variable, terminal, immediate, unequal or list
; variable: 'symbol
; terminal: number>0, denoting the nth terminal in rule
; immediate: "string"
; unequal: '(unequal expression) [not allowed on defining positions]
; list: '(list expressions...) [not allowed on matching positions]
;
; the output is a semantic program tree with nodes of the form:
; (node-name (attributes) list of nodes)


(load "pretty-print.so")
(define debug-grammar #t)   ;do a check for validity of rules

(define write-error
  (lambda (class text)
    (write
      (list
        class
        (list
          (cons 'source "source")
          '(stage eval))
        text)
      (current-error-port))
    (newline (current-error-port))
    #f))

; convert symbol or number to string
(define symbol-number->string
  (lambda (any)
    (if (symbol? any)
      (symbol->string any)
      (if (number? any)
        (number->string any)
        "<???>"))))

; is the non-terminal item a predicate rather than an alternatives?
(define not-terminal-is-predicate?
  (lambda (item)
    (null? (cddr item))))   ;nothing but actual params and predicate's name

; is the item a terminal ?
(define item-terminal?
  (lambda (item)
    (and
      (list? item)   ;pattern
      (symbol? (car item)))))   ;terminal-name a symbol

;;; generate an unique string value
(define generate-unique-value
  (lambda ()
    (set! global-unique-value (+ global-unique-value 1))
    global-unique-value))
(define global-unique-value 0)


;;; delete a list of variables from a list of needed ones
(define nolonger-needed
  (lambda (variables needed)
    (nolonger-needed-do variables needed '())))
(define nolonger-needed-do
  (lambda (variables old new)
    (if (null? old)
      new
      (nolonger-needed-do
        variables
        (cdr old)
        (if (member (car old) variables)
          new
          (cons (car old) new))))))


;;; determine whether all vars needed for the parameters are defined
; given is the list of yet defined variables
(define variables-athand?
  (lambda (params vars)
    (or
      (null? params)
      (and
        (variables-athand-do (car params) vars)
        (variables-athand? (cdr params) vars)))))
(define variables-athand-do
  (lambda (param vars)
    (or
      (null? param)   ;all param done without failure: good!
      (and
        (let ((item (car param)))
          (or
            (string? item)   ;terminal: ok
            (if
              (list? item)   ;sublist: check that
              (variables-athand-do (cdr item) vars)
              (assoc item vars))))   ;variable: ok, if in vars list
        (variables-athand-do (cdr param) vars)))))   ;go on in params

;;; determine whether all vars needed in the 'unequal subtree for the
; matching parameters are defined.
(define variables-unequal-athand?
  (lambda (params vars)
    (or
      (null? params)
      (and
        (variables-unequal-athand-do (car params) vars)
        (variables-unequal-athand? (cdr params) vars)))))
(define variables-unequal-athand-do
  (lambda (param vars)
    (or
      (null? param)   ;all done: good!
      (and
        (let ((item (car param)))
          (or
            (not (list? item))
            (if (null? item)
              (write-error
                'config
                "empty list not allowed during preparation")
              (if (equal? (car item) 'unequal)
                (variables-athand-do (cdr item) vars)
                (if (equal? (car item) 'list)
                  (variables-unequal-athand-do (cdr item) vars)
                  (write-error
                    'config
                    (string-append
                      "invalid list type during preparation: "
                      (symbol-number->string (car item)))))))))
        (variables-unequal-athand-do (cdr param) vars)))))


;;; determine the list of needed variables to evaluate the parameters
; given is the list of yet defined variables (thus not needed)
(define needed-variables
  (lambda (params vars)
    (needed-variables-do params vars '())))
(define needed-variables-do
  (lambda (params vars found)
    (if (null? params)   ;all params collect: return result
      found
      (needed-variables-do   ;collect further...
        (cdr params)   ;...together with the rest...
        vars
        (needed-variable (car params) vars found)))))
(define needed-variable
  (lambda (param vars found)
    (if (null? param)   ;parameter complete
      found
      (needed-variable
        (cdr param)
        vars
        (let ((p (car param)))
          (if (string? p)   ;terminal: nothing to add
            found
            (if (list? p)   ;sublist: collect all in it
              (needed-variable (cdr p) vars found)
              (if
                (or   ;variable: yet in vars or yet found -> nothing to add
                  (assoc p vars)
                  (member p found))
                found
                (cons p found)))))))))   ;...else add it

;;; apply variables to defining parameters and return parameter values
; if vars=#f, return #f
(define apply-variables
  (lambda (vars params pos)
    (and
      vars   ;if vars undefined: #f
      (apply-variables-do vars params pos '()))))
(define apply-variables-do
  (lambda (vars params pos values)
    (if (null? params)
      values
      (apply-variables-do
        vars
        (cdr params)
        pos
        (append
          values
          (list (apply-variable vars (car params) pos '())))))))
(define apply-variable
  (lambda (vars param pos value)
    (if (null? param)
      value
      (apply-variable
        vars
        (cdr param)
        pos
        (append
          value
          (let ((expr (car param)))
            (if (string? expr)
              (list expr)
              (if (list? expr)
                (if (null? expr)
                  (write-error
                    'config
                    "empty list not allowed during apply")
                  (if (equal? (car expr) 'list)
                    (list (apply-variable vars (cdr expr) pos '()))
                    (if (equal? (car expr) 'unique)
                      (list (generate-unique-value))
                      (if (equal? (car expr) 'source)
                        (list (list 'source pos))
                        (write-error
                          'config
                          (string-append
                            "invalid list type in defining position: "
                            (symbol-number->string (car expr))))))))
                (let ((v (assoc expr vars)))
                  (if v
                    (cdr v)
                    (write-error
                      'config
                      (string-append
                        "variable undefined in defining position: "
                        (symbol-number->string expr)))))))))))))

; variable: 'symbol or number [where a number denotes the nth terminal in rule]
; immediate: "string"
; list: '(list expressions...)
; unequal: '(unequal expression) [not allowed on defining positions]
; unique number: '(unique) [not allowed on matching positions]
; source position: '(source) [not allowed on matching positions]

;;; match parameter values against matching parameters and extract variables
; on failure, or if values=#f, return #f
(define match-variables
  (lambda (values params vars)
    (if (not vars)
      (write-error
        'internal
        "vars #f in match-variables"))
    (and
      values
      (match-variables-do values params vars '()))))
(define match-variables-do
  (lambda (values params vars news)
    (and
      news
      (if (null? values)
        (if (null? params)
          news
          (write-error
            'config
            "more values than params during match"))
        (if (null? params)
          (write-error
            'config
            "more params than values during match")
          (match-variables-do
            (cdr values)
            (cdr params)
            vars
            (match-variables-param
              (car values)
              (car params)
              vars
              news)))))))
(define match-variables-param
  (lambda (value param vars news)
    (and   ;news #f means prior call failed -> abort
      news
      (if (null? param)   ;parameter sequence done
        (and
          (null? value)   ;dangling values -> bad, #f
          news)
        (let ((item (car param)))
          (if (string? item)   ;constant, must match...
            (and
              (not (null? value))   ;...exactly one value...
              (string? (car value))   ;...which is a string...
              (equal? (car value) item)   ;...and is equal
              (match-variables-param
                (cdr value)   ;matched? so consume one of each!
                (cdr param)
                vars
                news))
            (if (list? item)
              (if (null? item)
                (write-error
                  'config
                  "empty list not allowed during match")
                (if (equal? (car item) 'list)   ;a '(list ...) matches...
                  (and
                    (not (null? value))
                    (list? (car value))   ;...a list (sic!)
                    (match-variables-param
                      (cdr value)
                      (cdr param)
                      vars
                      (match-variables-param   ;now match this sublist
                        (car value)
                        (cdar param)
                        vars
                        news)))
                  (if (equal? (car item) 'unequal)
                    (match-variables-unequality
                      value
                      (cdr item)
                      (cdr param)
                      vars
                      news)
                    (write-error
                      'config
                      (string-append
                        "invalid list type during match: "
                        (symbol-number->string (car item)))))))
              (let ((v (or (assoc item vars) (assoc item news))))
                (if v   ;if the variables is defined yet -> must be equal
                  (match-variables-equality
                    value
                    (cdr v)
                    (cdr param)
                    vars
                    news)
                  (if (null? value)   ;value exhausted: match against '() only
                    (match-variables-var
                      '()
                      item
                      '()
                      (cdr param)
                      vars
                      news)
                    (if (null? (cdr param))   ;if variable on last position...
                      (match-variables-var
                        value   ;match the whole rest against it
                        item
                        '()
                        '()
                        vars
                        news)
                      (match-variables-var   ;else just match one value-item.
                        (list (car value))
                        item
                        (cdr value)
                        (cdr param)
                        vars
                        news))))))))))))
; consume from value as much as needed and test for unequality.
; apply the variables and check them one by one, at least one must be unequal
(define match-variables-unequality
  (lambda (value unequal-item param vars news)
    (match-variables-unequal-do
      value
      (apply-variable (append vars news) unequal-item pos '())
      param
      vars
      news
      #t)))
; consume from value as much as needed and test for equality.
; variables are yet applied, check them one by one, all must be unequal
(define match-variables-equality
  (lambda (value equal-item param vars news)
    (match-variables-unequal-do
      value
      equal-item
      param
      vars
      news
      #f)))
(define match-variables-unequal-do
  (lambda (value refer param vars news shalldiffer)
    (if (null? refer)   ;all compared -> equality!
      (and
        (not shalldiffer)   ;too bad if them should differ -> #f
        (match-variables-param   ;...else proceed normally with next item
          value
          param
          vars
          news))
      (if (null? value)   ;value exhausted -> sure differs!
        (and
          shalldiffer   ;if should be equal -> #f
          (match-variables-param   ;else proceed normally
            '()
            param
            vars
            news))
        (if (equal? (car value) (car refer))   ;if next items are equal...
          (match-variables-unequal-do   ;...continue with the next ones
            (cdr value)
            (cdr refer)
            param
            vars
            news
            shalldiffer)
          (and   ;if items are not equal...
            shalldiffer   ;...failed, if should be equal -> #f
            (match-variables-unequal-skip   ;...else skip values, yet differs
              (cdr value)
              (cdr refer)
              param
              vars
              news)))))))
(define match-variables-unequal-skip
  (lambda (value refer param vars news)
    (if (null? refer)
      (match-variables-param
        value
        param
        vars
        news)
      (match-variables-unequal-skip
        (if (null? value)
          '()
          (cdr value))
        (cdr refer)
        param
        vars
        news))))
; add item with value, then do the rest.
(define match-variables-var
  (lambda (value item value-rest param vars news)
    (if (symbol? item)
      (match-variables-param
        value-rest
        param
        vars
        (cons   ;add the new variable with its matching value
          (cons item value)
          news))
      (write-error
        'config
        (string-append
          "cannot match non-symbol: "
          (if (number? item)
            (number->string item)
            "<???>"))))))

;rekursiv. in param:
;variable, definiert: matche die werte, failure->#f
;nummer: dito (muss definiert sein, sonst failure->#f)
;string: matche einzeln.
;(list..): matche rekursiv gegen eine liste; keine liste->#f
;(unequal..): matche gegen definiert werte. undefiniert->#f
;       i.e.: darf weder (unequal..) noch undefinierte variablen enthalten
;       frage also: muss in needed-variables etc auch der unequal-inhalt
;                   vom matchenden anteil geprueft werden?
;variable v, undefiniert:
;   in endposition: v := der ganze rest
;   ansonsten: v := ein element, egal was danach kommt.
;      (oder: nimm alles, und kuerze je um eines, bis der nachfolgende
;             Teil nicht mehr failed. Aber: Das ist _hoher_ Aufwand. Noetig?
;             offene Frage: anfangen mit allem, oder einem, oder keinem?)


;;; evaluate all elements of a predicate rule
; return the list of defined variables, or #f on failure
(define eval-pred-sub
  (lambda (predicates name sequence pos vars needed unevaluated)
    (if (null? needed)
      vars   ;now the variables are complete: return them
      (and
        (not (null? sequence))   ;if not resolvable: #f
        (let ((item (car sequence)))
          (if
            (and   ;if all variables needed are at hand:
              (variables-athand? (caar item) vars)   ;all in actual-in and...
                ;all in actual-out which are in an 'unequal subtree:
              (variables-unequal-athand? (cadar item) vars))
            (eval-pred-sub-vars   ;complete: eval this rule,then...
              predicates
              name
              (append unevaluated (cdr sequence))   ;...restart from beginning
              pos
              (eval-predicate
                predicates
                predicates
                (car item)
                (cadr item)
                pos
                vars)
              vars
              needed)
            (eval-pred-sub   ;cannot evaluate this due to lack of variables:
              predicates
              name
              (cdr sequence)   ;try the next one and...
              pos
              vars
              needed
              (append unevaluated (list item)))))))))   ;...postpone this item
(define eval-pred-sub-vars
  (lambda (predicates name sequence pos newvars vars needed)
    (and
      newvars   ;if evaluation of subexpression failed: #f
      (if (null? newvars)
        (eval-pred-sub predicates name sequence pos vars needed '())
        (if (assoc (caar newvars) vars)
          (write-error
            'config
            (string-append
              "variable defined twice in predicate: "
              (symbol-number->string (caar newvars))))
; ok? if both definitions are equal!
          (eval-pred-sub-vars
            predicates
            name
            sequence
            pos
            (cdr newvars)
            (cons (car newvars) vars)
            (nolonger-needed (list (caar newvars)) needed)))))))

;;; check node for valid structure
(define valid-node?
  (lambda (node)
    (and
      (list? node)
      (not (null? node))
      (symbol? (car node))   ;rule name
      (list? (cdr node))
      (not (null? (cdr node)))
      (list (cadr node))   ;formal parameters
      (not (null? (cadr node)))
      (list (caadr node))   ;formal input parameters
      (list (cdadr node))
      (not (null? (cdadr node)))
      (list (cadadr node))   ;formal output parameters
      (list (cddr node)))))   ;subtree sequence

;;; evaluate a predicate rule
; return output parameter values, or #f on failure
(define eval-pred-into
  (lambda (predicates param-in rule pos)
    (if (valid-node? rule)
      (eval-pred-into-do
        predicates
        param-in
        (car rule)   ;rule name
        (caadr rule)   ;formal input parameters
        (cadadr rule)   ;formal output parameters
        (cddr rule)   ;subtree sequence
        pos)   ;source position
      (write-error
        'config
        "invalid node structure at predicate"))))
(define eval-pred-into-do
  (lambda (predicates param-in name formal-in formal-out sequence pos)
    (let ((vars (match-variables param-in formal-in '())))
      (and
        vars   ;if couldnt match: #f
        (apply-variables   ;else return output parameter values after...
          (eval-pred-sub   ;...evaluating all the rule and applying the vars...
            predicates
            name
            sequence
            pos
            vars
            (needed-variables formal-out vars)
            '())
          formal-out   ;...to the formal output parameters
          pos)))))

;;; evaluate a predicate, with all needed variables given
; return the list of newly defined variables, or #f on failure
(define eval-predicate
  (lambda (predicates rest params name pos vars)
    (and
      (not (null? rest))   ;predicate not found -> #f
      (or
        (and   ;either it's the rule, and it matches!..
          (equal? name (caar rest))   ;check the rule's name
          (match-variables
            (eval-pred-into   ;try to eval this instance of the predicate
              predicates
              (apply-variables   ;...variables applied to...
                vars
                (car params)   ;...the actual input parameters...
                pos)
              (car rest)
              pos)
            (cadr params)   ;...and match against actual output parameters
            vars))
        (eval-predicate   ;..or next
          predicates
          (cdr rest)
          params
          name
          pos
          vars)))))

;;; evaluate a rule, with all needed variables given
; return the list of newly defined variables
(define eval-rule
  (lambda (predicates item vars)
    (match-variables
      (eval-tree   ;evaluate the item with...
        predicates 
        (apply-variables   ;...variables applied to...
          vars
          (caar item)   ;...the actual input parameters...
          (cadadr item))   ;source position
        (cdr item))
      (cadar item)   ;...and match against actual output parameters
      vars)))

;;; evaluate a subtree, with all entrant variables given
; as long as variables are needed to be evaluated
; try to evaluate the first element of the subtree sequence, proceed
; returns the list of variables
(define eval-subtree
  (lambda (predicates name sequence vars needed unevaluated)
    (if (null? sequence)   ;all items checked:
      (if (null? needed)
        vars   ;the variables are complete: return them
        (write-error   ;otherwise: without success!
          'config
          (string-append
            "rule unsolvable: "
            (symbol->string name)
            ", parameter: "
            (symbol-number->string (car needed)))))
      (let ((item (car sequence)))
        (if (item-terminal? item)
          (eval-subtree   ;terminal yet taken: extract and...
            predicates
            name
            (cdr sequence)   ;...proceed with next item
            vars
            needed
            unevaluated)
          (if
            (and   ;if all variables needed are at hand:
              (variables-athand? (caar item) vars)   ;all in actual-in and...
                ;all in actual-out which are in an 'unequal subtree:
              (variables-unequal-athand? (cadar item) vars))
            (eval-subtree-vars   ;complete: eval this rule, then...
              predicates
              name
              (append unevaluated (cdr sequence))   ;...restart from beginning
              (if (not-terminal-is-predicate? item)
                (eval-predicate
                  predicates
                  predicates
                  (car item)   ;parameters
                  (caadr item)   ;name
                  (cadadr item)   ;source position
                  vars)
                (eval-rule predicates item vars))
              vars
              needed)
            (eval-subtree   ;cannot evaluate this due to lack of variables:
              predicates
              name
              (cdr sequence)   ;try the next one and...
              vars
              needed
              (append
                unevaluated
                (list item)))))))))   ;...postpone this item
(define eval-subtree-vars
  (lambda (predicates name sequence newvars vars needed)
    (and
      newvars   ;newvars #f -> failed
      (if (null? newvars)
        (eval-subtree predicates name sequence vars needed '())
        (if (assoc (caar newvars) vars)
          (write-error
            'config
            (string-append
              "variable defined twice in non-predicate: "
              (symbol-number->string (caar newvars))))
; ok? if both definitions are equal! (only for predicates?)
          (eval-subtree-vars
            predicates
            name
            sequence
            (cdr newvars)
            (cons (car newvars) vars)
            (nolonger-needed (list (caar newvars)) needed)))))))

;;; extract all terminals and assign them to their index
(define term-variables
  (lambda (sequence vars)
    (term-variables-do sequence vars 1)))
(define term-variables-do
  (lambda (sequence vars index)
    (if (null? sequence)
      vars
      (term-variables-do
        (cdr sequence)
        (let ((item (car sequence)))
          (if (item-terminal? item)
            (cons (cons index item) vars)
            vars))
        (+ index 1)))))

;;; evaluate a tree, with input parameter values given
; match the input parameter values against the formal input params,
; determine the variables needed to fulfil the final variable insertion,
; eval the subtree,
; evaluate the formal output params by replacing the variable placeholders
; returns the output parameter values
(define eval-tree
  (lambda (predicates param-in tree)
    (eval-tree-do
      predicates
      param-in
      (caar tree)   ;rule name
      (cadar tree)   ;source position
      (caadr tree)   ;formal input parameters
      (cadadr tree)   ;formal output parameters
      (cddr tree))))   ;subtree sequence
(define eval-tree-do
  (lambda (predicates param-in name pos formal-in formal-out sequence)
    (let ((terms (term-variables sequence '())))
      (let ((vars (append terms (match-variables param-in formal-in terms))))
        (apply-variables
          (eval-subtree
            predicates
            name
            sequence
            vars
            (needed-variables formal-out vars)
            '())
          formal-out
          pos)))))

;;; check tree for validity, add file position information for each node,
; put it next to the rule name, thus replacing the cadr name with
; (name (y x ...))
(define filepos-and-valid?
  (lambda (tree pos)
    (and
      (valid-node? tree)
      (let ((newpos (filepos-sequence (cddr tree) pos)))
        (set-car! tree (list (car tree) newpos))
        newpos))))
(define filepos-sequence
  (lambda (sequence pos)
    (if (null? sequence)
      pos
      (let ((item (car sequence)))
        (let ((newpos
            (if (item-terminal? item)
              (caddr item)
              (if (not-terminal-is-predicate? item)
                (begin
                  (set-car! (cdr item) (list (cadr item) pos))
                  pos)
                (filepos-and-valid? (cdr item) pos)))))
          (and
            newpos
            (filepos-sequence (cdr sequence) newpos)))))))

;;; evaluate a full compilation unit
(define eval-file
  (lambda (predicates tree)
    (if tree   ;if the previous stage failed: just do nothing.
      (if (filepos-and-valid? tree '(0 0))
        (if (null? (cadadr tree))
          (write-error
            'config
            "missing output parameter at initial rule")
          (if (null? (cdr (cadadr tree)))
            (if (null? (caadr tree))
              (let ((result
                (eval-tree
                  predicates   ;the predicates go everywhere
                  '()   ;no input parameters for the full compilation unit
                  tree)))
                (and   ;unless an error occured: there is only one result -> car
                  result
                  (car result)))
              (write-error
                'config
                "input parameter at initial rule not allowed"))
            (write-error
              'config
              "more than one output parameter at initial rule")))
        (write-error
          'config
          "invalid node structure"))
      #f)))


;;; transform all strings in car position to symbols
(define car-string->symbol
  (lambda (tree)
    (if
      (and
        (list? tree)
        (not (null? tree)))
      (if (string? (car tree))
        (cons
          (string->symbol (car tree))
          (map car-string->symbol (cdr tree)))
        (map car-string->symbol tree))
      tree)))


(define predicates (make-parameter #f))
(load (cadr (argv)))
(pretty-print (car-string->symbol (eval-file (predicates) (read))))
(newline)

