;;;; mother modular generic compiler project
; (c)2001,2002/GPL Oskar Schirmer <oskar@scara.com>
; see file COPYING for GPL license details

; scan.scm: scanner
; usage:
;   scan finite-automaton-file <input-file >output-file
;
; the finite-automaton contains a list of
; state definitions as follows:
;
; (state
;   (chars       put-char-to-current-string?
;                    get-next-char-from-input?
;                        symbol-starts-here?
;                            next-state
;                                    symbol-to-produce)
;   ...
; )
; ...
; chars may be a char, a range (pair), a list, "else".
; put-char-to-current-string? may be a specific char to put,
;   or (usually), just #t to put the current character, or #f.
;   it also may be "up" to turn a character to uppercase, if applicable.
; the first state is always taken as the initial one.
; if next-state is a list, all but the first are pushed onto the state stack.
; if next-state is #f, the next state is popped off the state stack.
; scanning stops, when the state stack is empty at <eof>.
;
; the result of scanning are tokens represented as three element
; lists '(symbol "chars-collected-by-put" file-position)

(define write-error
  (lambda (class source text)
    (write
      (list
        class
        (list
          (cons 'source source)
          '(stage scan))
        text)
      (current-error-port))
    (newline (current-error-port))
    (write #t)))
    
;;; check for start of a word-found
(define trans-start
  (lambda (fina states input look-ahead read-ahead yx word-found word-pos finr)
    (if (caddr finr)   ;symbol_starts_here?
      (trans-put
        fina
        states
        input
        look-ahead
        read-ahead
        yx
        ""
        yx
        finr)
      (trans-put
        fina
        states
        input
        look-ahead
        read-ahead
        yx
        word-found
        word-pos
        finr))))

;;; append a char to the word-found
(define trans-put
  (lambda (fina states input look-ahead read-ahead yx word-found word-pos finr)
    (let ((put-this-one (car finr)))   ;put_char_to_current_string?
      (trans-symbol fina states input look-ahead read-ahead yx
        (if put-this-one
          (string-append word-found
            (if (eqv? put-this-one #t)
              (if (eof-object? input)
                "<EOF>"
                (string input))
              (if (eqv? put-this-one 'up)
                (if (eof-object? input)
                  "<EOF>"
                  (string (char-upcase input)))
                (string put-this-one))))
          word-found)
        word-pos
        finr))))

;;; emit a symbol
(define trans-symbol
  (lambda (fina states input look-ahead read-ahead yx word-found word-pos finr)
    (let ((sym (car (cddddr finr))))   ;symbol_to_produce
      (if
        (and   ;fatal error:
          (string? sym)   ;string instead of symbol...
          (not (cadddr finr)))   ;...and no next state given
        (write-error
          'fatal
          yx
          sym)
        (begin
          (if sym
            (if (string? sym)
              (write-error   ;normal error: string instead of symbol.
                'error
                yx
                sym)
              (begin
                (write
                  (list sym word-found word-pos))
                (newline))))
          (trans-next
            fina
            states
            input
            look-ahead
            read-ahead
            yx
            word-found
            word-pos
            finr))))))

;;; determine the next state
(define trans-next
  (lambda (fina states input look-ahead read-ahead yx word-found word-pos finr)
    (let ((st (cadddr finr)))
      (trans-stop
        fina
        (if st
          (if (list? st)
            (append st states)
            (cons st states))
          states)
        input
        look-ahead
        read-ahead
        yx
        word-found
        word-pos
        finr))))

;;; if the state-stack is empty: stop
(define trans-stop
  (lambda (fina states input look-ahead read-ahead yx word-found word-pos finr)
    (if (null? states)
      (if (eof-object? input)
        (write #f)
        (write-error
          'fatal
          yx
          (string-append
            "unknown character ("
            (number->string (char->integer input))
            ","
            (string input)
            ")")))
      (trans-get
        fina
        states
        input
        look-ahead
        read-ahead
        yx
        word-found
        word-pos
        finr))))

;;; check whether to get the next char
(define trans-get
  (lambda (fina states input look-ahead read-ahead yx word-found word-pos finr)
    (if (cadr finr)   ;get_next_char_from_input?
      (if (eqv? (cadr finr) 'back)
        (let ((ra (append look-ahead (list input) read-ahead)))
          (scan-state fina states (car ra) '() (cdr ra) yx word-found word-pos))
        (if (eof-object? input)
          (write-error 'fatal yx "unexpected end of file")
          (trans-read fina states
            (if (eqv? (cadr finr) 'look)
              (append look-ahead (list input))
              look-ahead)
            read-ahead yx word-found word-pos finr)))
      (scan-state
        fina
        states
        input
        look-ahead
        read-ahead
        yx
        word-found
        word-pos))))

;;; read a character
(define trans-read
  (lambda (fina states look-ahead read-ahead yx word-found word-pos finr)
    (if (null? read-ahead)
      (trans-line
        fina
        states
        (read-char)
        look-ahead
        read-ahead
        yx
        word-found
        word-pos
        finr)
      (trans-line
        fina
        states
        (car read-ahead)
        look-ahead
        (cdr read-ahead)
        yx
        word-found
        word-pos
        finr))))

;;; adjust the file position
(define trans-line
  (lambda (fina states input look-ahead read-ahead yx word-found word-pos finr)
    (scan-state fina states input look-ahead read-ahead
      (if (eqv? input #\newline)
        (list (+ (car yx) 1) 0)
        (list (car yx) (+ (cadr yx) 1)))
      word-found
      word-pos)))

;;; match the current input character with a state in the finite-automaton
(define scan-input
  (lambda (fina states input look-ahead read-ahead yx word-found word-pos fini)
    (if (null? fini)
      (write-error
        'fatal
        yx
        (string-append
          "unknown character ("
          (if (eof-object? input)
            "<EOF>"
            (string-append
              (number->string (char->integer input))
              ","
              (string input)))
          ")"))
      (let ((expect_char (caar fini)))
        (if
          (or
            (eqv? 'else expect_char)
            (if (list? expect_char)
              (memv input expect_char)
              (if (pair? expect_char)
                (and
                  (char? input)
                  (char>=? input (car expect_char))
                  (char<=? input (cdr expect_char)))
                (eqv? input expect_char))))
          (trans-start
            fina
            states
            input
            look-ahead
            read-ahead
            yx
            word-found
            word-pos
            (cdar fini))
          (scan-input
            fina
            states
            input
            look-ahead
            read-ahead
            yx
            word-found
            word-pos
            (cdr fini)))))))

;;; find the current state in the finite-automaton
(define scan-state
  (lambda (fina states input look-ahead read-ahead yx word-found word-pos)
    (let ((s (assq (car states) fina)))
      (if s
        (scan-input
          fina
          (cdr states)
          input
          look-ahead
          read-ahead
          yx
          word-found
          word-pos
          (cdr s))
        (write-error
          'config
          yx
          (string-append
            "undeclared state in finite automaton ("
            (symbol->string (car states))
            ")"))))))

(define finite-automaton (make-parameter #f))
(load (cadr (argv)))
(scan-state
  (finite-automaton)   ;the finite automaton
  (list (caar (finite-automaton)))   ;the first defined state
  (read-char)   ;the current input char
  '()   ;look-ahead
  '()   ;read-ahead
  '(1 1)   ;first line, first character
  ""   ;word-found
  '(0 0))   ;word-pos (none yet)
(newline)
