;;;; mother modular generic compiler project
; (c)2003/GPL Oskar Schirmer <oskar@scara.com>
; see file COPYING for GPL license details

; scope.scm: scope checker
; usage:
;   scope <input-file >output-file
;
; input and output are semantic program trees with nodes of the form:
; (node-name (attributes) list of nodes)


(load "pretty-print.so")
(load "srfi-1.so")

(define write-error
  (lambda (class text)
    (write
      (list
        class
        (list
          (cons 'source "source")
          '(stage scope))
        text)
      (current-error-port))
    (newline (current-error-port))
    #f))

;;; see whether an attribute is present at a node and return it,
; otherwise return #f
(define attribute?
  (lambda (id node)
    (let ((attrlist (cadr node)))
      (find (lambda (x) (equal? (car x) id)) attrlist))))

;;; check knowledge about a name
; return know/knowtype node if found, #f otherwise
(define known-by-name
  (lambda (name knowledge memory)
    (let ((node (assoc name knowledge)))
      (if node
        (caddr node)
        (and
          (not (null? memory))
          (known-by-name name (car memory) (cdr know-stack)))))))

;;; check knowledge about a ref
; return know/knowtype node if found, #f otherwise
(define known-by-ref
  (lambda (ref knowledge memory)
    (let ((node (find (lambda (x) (equal? (cadr x) ref)) knowledge)))
      (if node
        (caddr node)
        (and
          (not (null? memory))
          (known-by-ref ref (car memory) (cdr know-stack)))))))

;;; merge new attributes into an existing node
; if supersede is true then prefer the new attributes on equalness,
; else the old ones
(define merge-attributes
  (lambda (node attrlist supersede)
    (let ((merged
        (if supersede
          (merge-attrlist attrlist (cadr node))
          (merge-attrlist (cadr node) attrlist))))
      (and
        merged
        (cons
          (car node)
          (cons
            merged
            (cddr node)))))))
(define merge-attrlist
  (lambda (oldlist newlist)
    (if (null? newlist)
      oldlist
      (let ((same (assoc (caar newlist) oldlist)))   ;attr found in both?
        (if same
          (merge-attrlist oldlist (cdr newlist))
          (merge-attrlist
            (cons (car newlist) oldlist)
            (cdr newlist)))))))

;;; check all subnodes in a node, and the attribute type, too
; return the checked node, or #f on error
(define scope-subnodes
  (lambda (node knowledge memory istype)
    (let ((typeattr (attribute? 'type node)))
      (scope-subnodes-list
        (if typeattr
          (call-with-values
            (lambda ()
              (scope-node (cadr typeattr) knowledge memory #t))
            (lambda (newnode newknowledge)
              (and
                newnode
                (merge-attributes node (list (list 'type newnode)) #t))))
          node)
        knowledge
        memory
        istype))))
(define scope-subnodes-list
  (lambda (node knowledge memory istype)
    (call-with-values
      (lambda ()
        (scope-list (cddr node) '() knowledge memory istype))
      (lambda (checked newknowledge)
        (values
          (cons
            (car node)
            (cons
              (cadr node)
              checked))
          newknowledge)))))

;;; scope check a 'choose node
; check all its subnodes, silently omitting the erroneous ones,
; return #f just in case none of the subnodes is valid
(define scope-choose
  (lambda (node knowledge memory istype)
    (let ((left
          (scope-choose-list (cddr node) '() knowledge memory istype)))
      (and   ;if nothing to choose: #f, otherwise...
        (not (null? left))
        (cons   ;...replace the sublist by what is left
          (car node)
          (cons
            (cadr node)
            left))))))
(define scope-choose-list
  (lambda (nodes left knowledge memory istype)
    (if (null? nodes)
      left
      (call-with-values
        (lambda ()
          (scope-node (car nodes) knowledge memory istype))
        (lambda (newnode newknowledge)
          (if newnode
            (if (eq? knowledge newknowledge)
              (scope-choose-list
                (cdr nodes)
                (append left newnode)
                knowledge
                memory
                istype)
              (write-error 'fatal "choose changed knowledge"))
            (scope-choose-list   ;node failed: it is not left
              (cdr nodes)
              left
              knowledge
              memory
              istype)))))))

;;; scope check a 'use node
; return the checked or replaced node, or #f on error
(define scope-use
  (lambda (node knowledge memory istype)
    (let ((nameattr (attribute? 'name node)))
      (and
        nameattr   ;node must contain 'name attribute
        (let ((knownode (known-by-name (cdr nameattr) knowledge memory)))
          (and
            knownode
            (equal? (car knownode) (if istype 'knowtype 'know))
            (if istype
              (let ((typeattr (attribute? 'type knownode)))
                (and
                  typeattr
                  (merge-attributes
                    (cadr typeattr)
                    (remove (lambda (x) (equal? (car x) 'type)) (cadr knownode))
                    #t)))
              (if (null? (cddr knownode))
                (merge-attributes node (cadr knownode) #f)
                (merge-attributes (caddr knownode) (cadr node) #t)))))))))

;;; scope check a node
; returns the scope-checked tree, or #f on error; and the knowledge
(define scope-node
  (lambda (node knowledge memory istype)
    (cond
      ( (or
          (equal? (car node) 'know)
          (equal? (car node) 'knowtype))
        (let ((nameattr (attribute? 'name node))
              (refattr (attribute? 'ref node)))
          (if (and nameattr refattr)
            (let ((name (cdr nameattr)))
              (if (known-by-name name knowledge '())
                (values
                  (write-error 'fatal "name known twice")
                  knowledge)
                (values
                  '()
                  (call-with-values
                    (lambda ()
                      (scope-subnodes node knowledge memory istype))
                    (lambda (newnode newknowledge)
                      (cons
                        (list
                          name
                          (cadr refattr)
                          newnode)
                        newknowledge))))))
            (values
              (write-error 'fatal "name or ref attribute missing to know")
              knowledge))))
      ( (or
          (equal? (car node) 'forget)
          (equal? (car node) 'forgettype))
        (let ((refattr (attribute? 'ref node)))
          (if refattr
            (let ((ref (cadr refattr)))
              (if (known-by-ref ref knowledge '())
                (values
                  '()
                  (remove (lambda (x) (equal? (cadr x) ref)) knowledge))
                (values
                  (write-error 'fatal "forget off scope")
                  knowledge)))
            (values
              (write-error 'fatal "ref attribute missing to forget")
              knowledge))))
      ( (equal? (car node) 'scope)
        (values
          (call-with-values
            (lambda ()
              (scope-list
                (cddr node)
                '()
                '()
                (cons knowledge memory)
                istype))
            (lambda (checkednode finalknowledge)
              (if (null? finalknowledge)
                (values
                  checkednode
                  knowledge)
                (values
                  (write-error 'fatal "unforgotten at end of scope")
                  '()))))))
      ( (equal? (car node) 'choose)
        (values
          (scope-choose node knowledge memory istype)
          knowledge))
      ( (equal? (car node) 'use)
        (values
          (scope-use node knowledge memory istype)
          knowledge))
      ( else
        (let ((refattr (attribute? 'ref node)))
          (if refattr
            (let ((ref (cadr refattr)))
              (let ((byref (known-by-ref ref knowledge memory)))
                (if byref
                  (call-with-values
                    (lambda ()
                      (scope-subnodes node knowledge memory istype))
                    (lambda (newnode newknowledge)
                      (values
                        (merge-attributes newnode (cadr byref) #f)
                        newknowledge)))
                  (values
                    (write-error 'fatal "unknown reference")
                    knowledge))))
            (scope-subnodes node knowledge memory istype)))))))


;;; scope check a list of nodes
; return the checked list, or #f on error
; and return the current knowledge
(define scope-list
  (lambda (unchecked checked knowledge memory istype)
    (if (null? unchecked)
      (values
        checked
        knowledge)
      (call-with-values
        (lambda ()
          (scope-node (car unchecked) knowledge memory istype))
        (lambda (checkednode newknowledge)
          (if checkednode
            (scope-list
              (cdr unchecked)
              (if (null? checkednode)
                checked   ;omit this node, it is obsolete now
                (append checked (list checkednode)))
              newknowledge
              memory
              istype)
            (values
              #f
              newknowledge)))))))

;;; scope check the full file node list
; return it checked
(define scope-file
  (lambda (nodelist)
    (call-with-values
      (lambda ()
        (scope-list nodelist '() '() '() #f))
      (lambda (nodelist knowledge)
        (if (null? knowledge)
          nodelist
          (write-error 'fatal "unforgotten at end of file"))))))

; (define predicates (make-parameter #f))
; (load (cadr (argv)))
(pretty-print (scope-file (read)))
(newline)
